///////////////////////////////////////////////////////////////////////////////
// File: PerformanceCounters.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "PerformanceCounters.h"
#include "PerfCountersApplicationName.h"

#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/Win32Tools/SystemUtils.h"
#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"

#include "JetByteTools/PerfMonTools/PerformanceCounterInstaller.h"
#include "JetByteTools/PerfMonTools/SharedMemoryMutexPerformanceMonitorCounters.h"
#include "JetByteTools/PerfMonTools/PerformanceDataSchemaFileExporter.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::CException;
using JetByteTools::Core::StringVector;

using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CSecurityAttributes;
using JetByteTools::Win32::GetModulePathName;

using JetByteTools::PerfMon::CPerformanceCounterInstaller;
using JetByteTools::PerfMon::CPerformanceDataSchemaFileExporter;
using JetByteTools::PerfMon::CSharedMemoryMutexPerformanceMonitorCounters;
using JetByteTools::PerfMon::CPerformanceMonitorCountersBase;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static CPerformanceMonitorCountersBase *CreateCountersIfRequired(
   bool createCounters,
   const _tstring &applicationName,
   const CPerformanceDataSchema &schema,
   const CSecurityAttributes &sa);

///////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////

static const CPerformanceMonitorCountersBase::ObjectId s_noObject;

static const CPerformanceMonitorCountersBase::CounterId s_noCounter;

///////////////////////////////////////////////////////////////////////////////
// CPerformanceCounters
///////////////////////////////////////////////////////////////////////////////

CPerformanceCounters::CPerformanceCounters(
   const bool createCounters,
   CSecurityAttributes &sa,
   const _tstring &logName,
   const _tstring &outputFileName)
   :  m_pCounters(CreateCountersIfRequired(createCounters, s_applicationName, m_schema, sa)),
      m_object(m_pCounters ? m_pCounters->GetPerformanceObject(m_schema.s_objectName) : s_noObject),
      m_bytesProcessedCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_bytesProcessedCounter) : s_noCounter),
      m_totalBuffersCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_totalBuffersCounter) : s_noCounter),
      m_buffersInUseCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_buffersInUseCounter) : s_noCounter),
      m_connectionsActiveCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_connectionsActiveCounter) : s_noCounter),
      m_totalSocketsCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_totalSocketsCounter) : s_noCounter),
      m_socketsInUseCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_socketsInUseCounter) : s_noCounter),
      m_ioThreadsActiveCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_ioThreadsActiveCounter) : s_noCounter),
      m_ioThreadsProcessingCounter(m_pCounters ? m_pCounters->GetPerformanceCounter(m_object, m_schema.s_ioThreadsProcessingCounter) : s_noCounter),
      m_logger(logName, *this, outputFileName)
{
}

CPerformanceCounters::~CPerformanceCounters()
{
   delete m_pCounters;
}

bool CPerformanceCounters::LogCountersWithOptionalAdditionalCounters(
   const _tstring &templateFile) const
{
   return m_logger.InstallAndStartWithOptionalAdditionalCounters(templateFile);
}

bool CPerformanceCounters::LogCountersWithOptionalAdditionalCounters(
   const _tstring &templateFile,
   const _tstring &parameters) const
{
   return m_logger.InstallAndStartWithOptionalAdditionalCounters(templateFile, parameters);
}

bool CPerformanceCounters::CountersAreActive() const
{
   return m_pCounters != nullptr;
}

const CPerformanceDataSchema &CPerformanceCounters::GetSchema() const
{
   return m_schema;
}

bool CPerformanceCounters::GetInstanceNamesForObject(
   const _tstring &objectName,
   StringVector &instanceNames) const
{
   if (m_pCounters)
   {
      return m_pCounters->GetInstanceNamesForObject(objectName, instanceNames);
   }

   return false;
}

bool CPerformanceCounters::InstallCounters()
{
   bool installed = false;

   CPerformanceCounterInstaller installer;

   if (!installer.CountersAreInstalled(s_applicationName))
   {
      const _tstring counterPath = GetModulePathName() + _T("\\EchoServerServicePerfCounters.dll");

      CPerformanceDataSchema schema;

      CPerformanceDataSchemaFileExporter exporter;

      installer.Install(
         s_applicationName,
         counterPath,
         false,
         exporter.ExportHeaderFile(s_applicationName, schema),
         exporter.ExportIniFile(s_applicationName, schema),
         _T(""),
         10000,
         10000,
         schema.GetChecksum());

      installed = true;
   }

   return installed;
}

bool CPerformanceCounters::RemoveCounters()
{
   bool removed = false;

   CPerformanceCounterInstaller installer;

   if (installer.CountersAreInstalled(s_applicationName))
   {
      installer.Uninstall(s_applicationName);

      removed = true;
   }

   return removed;
}

void CPerformanceCounters::OnSocketCreated()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_totalSocketsCounter);
   }
}

void CPerformanceCounters::OnSocketAttached(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_socketsInUseCounter);
   }
}

void CPerformanceCounters::OnConnectionStarted()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_connectionsActiveCounter);
   }
}

void CPerformanceCounters::OnConnectionComplete()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_connectionsActiveCounter);
   }
}

void CPerformanceCounters::OnSocketReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_socketsInUseCounter);
   }
}

void CPerformanceCounters::OnSocketDestroyed()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_totalSocketsCounter);
   }
}

void CPerformanceCounters::BytesProcessed(
   const long bytes)
{
   if (m_pCounters)
   {
      m_pCounters->AddToPerformanceCounter64(m_bytesProcessedCounter, bytes);
   }
}

void CPerformanceCounters::OnBufferCreated()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_totalBuffersCounter);
   }
}

void CPerformanceCounters::OnBufferDestroyed()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_totalBuffersCounter);
   }
}

void CPerformanceCounters::OnBufferAllocated(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_buffersInUseCounter);
   }
}

void CPerformanceCounters::OnBufferAllocationContention()
{
}

void CPerformanceCounters::OnBufferReleaseContention()
{
}

void CPerformanceCounters::OnBufferReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_buffersInUseCounter);
   }
}

void CPerformanceCounters::OnThreadPoolError(
   const _tstring & /*message*/)
{
}

void CPerformanceCounters::OnThreadPoolThreadCreated()
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_ioThreadsActiveCounter);
   }
}

void CPerformanceCounters::OnThreadPoolDispatch()
{
}

void CPerformanceCounters::OnThreadPoolThreadDestroyed()
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_ioThreadsActiveCounter);
   }
}

void CPerformanceCounters::OnThreadPoolProcessingStarted(
   const bool /*dataAvailable*/)
{
   if (m_pCounters)
   {
      m_pCounters->IncrementPerformanceCounter32(m_ioThreadsProcessingCounter);
   }
}

void CPerformanceCounters::OnThreadPoolProcessingStopped(
   const bool /*dataAvailable*/)
{
   if (m_pCounters)
   {
      m_pCounters->DecrementPerformanceCounter32(m_ioThreadsProcessingCounter);
   }
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static CPerformanceMonitorCountersBase *CreateCountersIfRequired(
   const bool createCounters,
   const _tstring &applicationName,
   const CPerformanceDataSchema &schema,
   const CSecurityAttributes &sa)
{
   CPerformanceMonitorCountersBase *pCounters = nullptr;

   if (createCounters)
   {
      if (CPerformanceCounterInstaller().CountersAreInstalled(applicationName))
      {
         if (CPerformanceCounterInstaller().GetSchemaChecksum(applicationName) != schema.GetChecksum())
         {
            throw CException(
               _T("CPerformanceCounters::CreateCountersIfRequired()"),
               _T("Performance counter schema mismatch; these counters aren't the ones that are installed."));
         }

         const CGlobalName name(applicationName);

         const DWORD firstCounterIndex = CPerformanceCounterInstaller().FirstCounterIndex(s_applicationName);

         pCounters = new CSharedMemoryMutexPerformanceMonitorCounters(name, schema, firstCounterIndex, const_cast<CSecurityAttributes *>(&sa));
      }
      else
      {
         //OutputEx(_T("Not using performance counters as they are not currently installed. Run the service with /InstallCounters to install them."));
      }
   }

   return pCounters;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: PerformanceCounters.cpp
///////////////////////////////////////////////////////////////////////////////
