///////////////////////////////////////////////////////////////////////////////
// File: PerformanceDataCollector.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2007 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "PerformanceDataCollector.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/SEHException.h"

#include "JetByteTools/PerfMonTools/PerformanceCounterInstaller.h"
#include "JetByteTools/PerfMonTools/SharedMemoryMutexPerformanceDataBlock.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;

using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CSEHException;

using JetByteTools::PerfMon::CPerformanceCounterInstaller;
using JetByteTools::PerfMon::IPerformanceDataBlock;
using JetByteTools::PerfMon::CSharedMemoryMutexPerformanceDataBlock;

///////////////////////////////////////////////////////////////////////////////
// CPerformanceDataCollector
///////////////////////////////////////////////////////////////////////////////

CPerformanceDataCollector::CPerformanceDataCollector(
   const _tstring &applicationName)
   :  m_name(applicationName),
      m_sa(m_sd),
      m_collector(
         *this,
         m_schema,
         CPerformanceCounterInstaller().FirstCounterIndex(applicationName))
{
}

IPerformanceDataBlock &CPerformanceDataCollector::AllocatePerformanceDataBlock(
   const JetByteTools::PerfMon::CPerformanceDataSchema &schema,
   const DWORD firstCounterIndex)
{
   IPerformanceDataBlock *pDataBlock = new CSharedMemoryMutexPerformanceDataBlock(m_name, schema, firstCounterIndex, &m_sa);

   return *pDataBlock;
}

void CPerformanceDataCollector::ReleasePerformanceDataBlock(
   IPerformanceDataBlock &dataBlock)
{
   (void)dataBlock;
}

void CPerformanceDataCollector::Open(
   const wchar_t * const pDevNames)
{
   CSEHException::Translator sehTranslator(0);

   try
   {
#ifdef _DEBUG
      const std::wstring message = std::wstring(L"CPerformanceDataCollector::Open() -  ") + (pDevNames ? pDevNames : L"");

      OutputDebugStringW(message.c_str());
#endif
      m_collector.Open(pDevNames);
   }
   catch(CException &e)
   {
      const _tstring message = _T("CPerformanceDataCollector::Open() -  Exception: ") + e.GetDetails();

      OutputDebugString(message.c_str());
   }
   catch(const std::exception &e)
   {
      const _tstring message = _T("CPerformanceDataCollector::Open() - std::exception - ") + CStringConverter::AtoT(e.what());

      OutputDebugString(message.c_str());
   }
   catch(...)
   {
      OutputDebugString(_T("CPerformanceDataCollector::Open() - Unexpected exception"));
   }

   DEBUG_ONLY(OutputDebugString(_T("CPerformanceDataCollector::Open() - Complete")));
}

void CPerformanceDataCollector::Close()
{
   CSEHException::Translator sehTranslator(0);

   try
   {
      DEBUG_ONLY(OutputDebugString(_T("CPerformanceDataCollector::Close()")));

      m_collector.Close();
   }
   catch(CException &e)
   {
      const _tstring message = _T("CPerformanceDataCollector::Close() -  Exception: ") + e.GetDetails();

      OutputDebugString(message.c_str());
   }
   catch(const std::exception &e)
   {
      const _tstring message = _T("CPerformanceDataCollector::Close() - std::exception - ") + CStringConverter::AtoT(e.what());

      OutputDebugString(message.c_str());
   }
   catch(...)
   {
      OutputDebugString(_T("CPerformanceDataCollector::Close() - Unexpected exception"));
   }

   DEBUG_ONLY(OutputDebugString(_T("CPerformanceDataCollector::Close() - Complete")));
}

bool CPerformanceDataCollector::CollectW(
   const wchar_t * const pszValueName,
   BYTE **ppData,
   DWORD *pcbTotalBytes,
   DWORD *pNumObjectTypes)
{
   bool ok = true;

   CSEHException::Translator sehTranslator(0);

   try
   {
#ifdef _DEBUG
      const std::wstring message = std::wstring(L"CPerformanceDataCollector::Collect() -  ") + (pszValueName ? pszValueName : L"");

      OutputDebugStringW(message.c_str());
#endif

      ok = m_collector.Collect(pszValueName, ppData, pcbTotalBytes, pNumObjectTypes);

#ifdef _DEBUG
      const _tstring results = _T("OK: ") + ToString(ok) + _T(" bytes: ") + ToString(*pcbTotalBytes) + _T(" objects: ") + ToString(*pNumObjectTypes);

      OutputDebugString(results.c_str());
#endif
   }
   catch(CException &e)
   {
      const _tstring message = _T("CPerformanceDataCollector::Collect() -  Exception: ") + e.GetDetails();

      OutputDebugString(message.c_str());
   }
   catch(const std::exception &e)
   {
      const _tstring message = _T("CPerformanceDataCollector::Collect() - std::exception - ") + CStringConverter::AtoT(e.what());

      OutputDebugString(message.c_str());
   }
   catch(...)
   {
      OutputDebugString(_T("CPerformanceDataCollector::Collect() - Unexpected exception"));
   }

   DEBUG_ONLY(OutputDebugString(_T("CPerformanceDataCollector::Collect() - Complete")));

   return ok;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: PerformanceDataCollector.cpp
///////////////////////////////////////////////////////////////////////////////
